package com.loafle.overflow.proxy;

import com.google.gson.internal.Primitives;
import com.google.protobuf.ByteString;
import org.codehaus.jackson.map.DeserializationConfig;
import org.codehaus.jackson.map.ObjectMapper;
import org.codehaus.jackson.type.JavaType;
import org.springframework.aop.support.AopUtils;
import org.springframework.context.ApplicationContext;
import org.springframework.stereotype.Component;

import java.io.IOException;
import java.lang.reflect.*;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Created by crusader on 17. 6. 30.
 */
@Component
public class ServiceInvoker {
//    @Autowired
    private ApplicationContext context;

    private ObjectMapper objectMapper;

    private Map<String, Cache> serviceCacheMap;

    public ServiceInvoker(ApplicationContext context) {
        this.context = context;
        objectMapper = new ObjectMapper();
        objectMapper.configure(DeserializationConfig.Feature.FAIL_ON_UNKNOWN_PROPERTIES, false);

        serviceCacheMap = new HashMap<>();
    }

    private static class Cache {
        private ServiceCache serviceCache;
        private Map<String, MethodCache> methodCacheMap;

        private Cache(Object bean) {
            serviceCache = new ServiceCache();
            serviceCache.bean = bean;

            methodCacheMap = new HashMap<>(10);
        }

        private Object getBean() {
            return serviceCache.bean;
        }

        private MethodCache getMethodCache(String methodName) throws NoSuchMethodException, ClassNotFoundException {
            MethodCache methodCache = methodCacheMap.get(methodName);
            if (null != methodCache) {
                return methodCache;
            }

            Method method = getMethod(methodName);
            if (null == method) {
                throw new NoSuchMethodException();
            }
            methodCache = new MethodCache();
            methodCacheMap.put(methodName, methodCache);

            methodCache.method = method;
            methodCache.returnClazz = method.getReturnType();

            int paramCount = method.getParameterCount();
            if (0 == paramCount) {
                return methodCache;
            }

            methodCache.parameterTypes = method.getGenericParameterTypes();

            return methodCache;
        }

        private Method getMethod(String methodName) {
            Class<?> clazz = AopUtils.getTargetClass(serviceCache.bean);
            Method[] methods = clazz.getMethods();

            Method targetMethod = null;
            for(Method m : methods){
                if (methodName.equals(m.getName())) {
                    targetMethod = m;
                    break;
                }
            }
            return targetMethod;
        }

        private static class ServiceCache {
            private Object bean;
            private Map<String, MethodCache> methodCacheMap;
        }
        private static class MethodCache {
            private Method method;
            private Class returnClazz;
            private Type[] parameterTypes;
        }
    }

    private Cache getServiceCache(String serviceName) {
        Cache serviceCache = serviceCacheMap.get(serviceName);
        if (null == serviceCache) {
            Object bean = context.getBean(serviceName);
            serviceCache = new Cache(bean);
            serviceCacheMap.put(serviceName, serviceCache);
        }

        return serviceCache;
    }

    private Object getValue(Type parameterType, String json) throws IOException {
        JavaType targetType = objectMapper.getTypeFactory().constructType(parameterType);
        if (!Primitives.isPrimitive(parameterType) && !parameterType.getTypeName().equals(String.class.getName())) {
            return objectMapper.readValue(json, targetType);
        }
        return objectMapper.convertValue(json, targetType);
    }


    private Object[] getParameters(Type[] parameterTypes, List<ByteString> params) throws IOException {
        if (null == parameterTypes || null == params) {
            return null;
        }

        if (parameterTypes.length != params.size()) {
            throw new IllegalArgumentException();
        }

        Object[] result = new Object[parameterTypes.length];

        for (int i = 0; i < parameterTypes.length; i++) {
            result[i] = getValue(parameterTypes[i], params.get(i).toStringUtf8());
        }

        return result;
    }


    public String invoke(String serviceName, String methodName, List<ByteString> params) throws NoSuchMethodException, ClassNotFoundException, IOException, InvocationTargetException, IllegalAccessException {
        Cache serviceCache = getServiceCache(serviceName);
        Cache.MethodCache methodCache = serviceCache.getMethodCache(methodName);

        Object[] parameters = null;
        if (null != methodCache.parameterTypes && 0 < methodCache.parameterTypes.length) {
            parameters = getParameters(methodCache.parameterTypes, params);
        }

        Object result = methodCache.method.invoke(serviceCache.getBean(), parameters);

        String jsonInString = objectMapper.writeValueAsString(result);

        return jsonInString;
    }
}
